import numpy as np
from .utils import se3Exp, se3Log
from scipy.interpolate import RegularGridInterpolator

def derive_jac_and_residual(img, depth, img_tag, xi, K):
    # compute the jacobian and residual corresponding to the given xi.

    ########################################################################
    # TODO: get the rotation and translation from the xi                   #
    ########################################################################

    T = se3Exp(xi)
    R = T[:3, :3]
    t = T[:3, 3]

    ########################################################################
    #                           END OF YOUR CODE                           #
    ########################################################################

    # the arries to save the intermediate results
    # these two contain the x, y, image coordinates of the reference-pixel, 
    # transformed & projected into the new image
    Imgx = np.zeros(img.shape) - 10
    Imgy = np.zeros(img.shape) - 10

    # these arraies save the 3d position of the transformed point
    xp = np.zeros(img.shape) + np.float64('nan')
    yp = np.zeros(img.shape) + np.float64('nan')
    zp = np.zeros(img.shape) + np.float64('nan')

    for i in range(img.shape[0]):
        for j in range(img.shape[1]):
    ########################################################################
    # TODO: 1) get the point in 3D space using the depth map               #
    # 2) rotate the point and translate it to target camera                # 
    # 3) if it is valid (depth >0 in both), project and save results       #
    ########################################################################

            p = np.linalg.inv(K) @ (np.array([j, i, 1]) * depth[i, j])

            pTrans = R @ p + t

            if pTrans[-1] > 0 and depth[i, j] > 0:
                pTransProj = K @ pTrans
                Imgx[i, j] = pTransProj[0] / pTransProj[2]
                Imgy[i, j] = pTransProj[1] / pTransProj[2]

                xp[i, j] = pTrans[0]
                yp[i, j] = pTrans[1]
                zp[i, j] = pTrans[2]
            
            
    ########################################################################
    #                           END OF YOUR CODE                           #
    ########################################################################

    # compute the image derivative
    dxI = np.zeros(img.shape) + np.float64('nan')
    dyI = np.zeros(img.shape) + np.float64('nan')
    dxI[:, 1:-1] = (img_tag[:, 2:] - img_tag[:, :-2]) / 2
    dyI[1:-1, :] = (img_tag[2:, :] - img_tag[:-2, :]) / 2
    interpx = RegularGridInterpolator((np.arange(img.shape[0]), np.arange(img.shape[1])), dxI, bounds_error=False)
    interpy = RegularGridInterpolator((np.arange(img.shape[0]), np.arange(img.shape[1])), dyI, bounds_error=False)
    Ixfx = K[0, 0] * interpx(np.stack([Imgy.flatten(), Imgx.flatten()], axis=1))
    Iyfy = K[1, 1] * interpy(np.stack([Imgy.flatten(), Imgx.flatten()], axis=1))

    # get warped 3d points
    xp = xp.flatten()
    yp = yp.flatten()
    zp = zp.flatten()
    
    # implement the gradient
    Jac = np.zeros((img.shape[0] * img.shape[1], 6))
    interpI = RegularGridInterpolator((np.arange(img.shape[0]), np.arange(img.shape[1])), img_tag, bounds_error=False)
    ########################################################################
    # TODO: compute the Jacobian and residual                              #
    # check the theoretical exericse to get the Jacobian formula           #
    ########################################################################

    Jac[:, 0] = Ixfx / zp
    Jac[:, 1] = Iyfy / zp
    Jac[:, 2] = -(Ixfx * xp + Iyfy * yp) / zp ** 2
    Jac[:, 3] = -(Ixfx * xp * yp) / zp **2 - Iyfy * (1 + (yp / zp)  ** 2)
    Jac[:, 4] = Ixfx * (1 + (xp / zp) ** 2) + Iyfy * xp * yp / zp ** 2
    Jac[:, 5] = (-Ixfx * yp + Iyfy * xp) / zp

    residual = interpI(np.stack([Imgy.flatten(), Imgx.flatten()], axis=1)) - img.flatten()

    ########################################################################
    #                           END OF YOUR CODE                           #
    ########################################################################
    
    # remove not valid pixels
    notvalid = np.isnan(np.sum(Jac, axis=1))
    residual[notvalid] = 0
    Jac[notvalid, :] = 0

    return Jac, residual


def update_step(jac, residual, xi):
    # compute the update step using the Gauss-Newton method
    # input: jac: the Jacobian J, num_pixel x 6
    #        residual: the residual r, num_pixel x 1
    #        xi: the current xi, 6 x 1
    # output: xi: the updated xi

    ########################################################################
    # TODO: compute the update step using the Gauss-Newton method          #
    # 1. compute the descent setp for \xi                                  #
    # 2. update the \xi using the descent step                             #
    ########################################################################

    upd = -np.linalg.inv(jac.T @ jac) @ jac.T @ residual

    xi = se3Log(se3Exp(upd) @ se3Exp(xi))

    ########################################################################
    #                           END OF YOUR CODE                           #
    ########################################################################

    return xi