import numpy as np
from PIL import Image
from scipy.linalg import expm, logm

def se3Exp(twist):
    # twist: 6x1 vector
    ########################################################################
    # TODO: Implement the function to compute the mat exp of the twist     #
    # construct the corresponding mat in se3                               #
    ########################################################################

    M = np.array([[0, -twist[5], twist[4], twist[0]],
                  [twist[5], 0, -twist[3], twist[1]],
                  [-twist[4], twist[3], 0, twist[2]],
                  [0, 0, 0, 0]])

    ########################################################################
    #                           END OF YOUR CODE                           #
    ########################################################################

    return expm(M)

def se3Log(T):
    # T: 4x4 matrix
    lg = logm(T)
    ########################################################################
    # TODO: Implement the function to compute the mat log of the SE3 mat   #
    # extract the twist vector from the corresponding se3 mat              #
    ########################################################################

    twist = np.array([lg[0, 3], lg[1, 3], lg[2, 3], lg[2, 1], lg[0, 2], lg[1, 0]])

    ########################################################################
    #                           END OF YOUR CODE                           #
    ########################################################################

    return twist


def load_dataset():
    K = np.array([[517.3, 0, 318.6],
                  [0, 516.5, 255.3],
                  [0, 0, 1]]) 
    img1 = np.array(Image.open('./rgb/1305031102.275326.png').convert('L')) / 255.
    img2 = np.array(Image.open('./rgb/1305031102.175304.png').convert('L')) / 255.
    depth1 = np.array(Image.open('./depth/1305031102.262886.png')) / 5000.
    depth2 = np.array(Image.open('./depth/1305031102.160407.png')) / 5000.

    return img1, img2, depth1, depth2, K

def down_scale(img, depth, K, lvl):
    # downscale the color image, depth map and camera intrisics given the level

    # if we are at the lowest level, we don't need to downscale
    if lvl == 1:
        return img, depth, K
    ########################################################################
    # TODO: Implement the downscale of the image, depth and camera         #
    # image is discretized at the exact pixel values (e.g. 3, 7)           #
    # not at the center of each pixel (e.g. 3.5, 7.5)                      #
    # note that the depth map should only consider the valid pixels        #
    ########################################################################

    K_downscale = np.array([[K[0, 0] /2 , 0, (K[0, 2] + 0.5) / 2 - 0.5],
                            [0, K[1, 1] /2, (K[1, 2] + 0.5) / 2 - 0.5],
                            [0, 0, 1]])
    
    img_downscale = 0.25 * (img[::2, ::2] + img[1::2, ::2] + img[::2, 1::2] + img[1::2, 1::2])

    depth_valid = (np.sign(depth[::2, ::2]) + np.sign(depth[1::2, ::2]) 
                   + np.sign(depth[::2, 1::2]) + np.sign(depth[1::2, 1::2]))
    depth_downscale = (depth[::2, ::2] + depth[1::2, ::2] + depth[::2, 1::2] + depth[1::2, 1::2]) / depth_valid 
    # the element that is nan set to 0
    depth_downscale[np.isnan(depth_downscale)] = 0

    ########################################################################
    #                           END OF YOUR CODE                           #
    ########################################################################
    
    # recursively downscale
    return down_scale(img_downscale, depth_downscale, K_downscale, lvl - 1)
    
